import 'package:hangman/models/settings/settings_game.dart';
import 'package:hangman/models/settings/settings_global.dart';
import 'package:hangman/utils/tools.dart';

typedef MovingTile = String;
typedef Player = String;
typedef ConflictsCount = List<List<int>>;
typedef AnimatedBoard = List<List<bool>>;
typedef AnimatedBoardSequence = List<AnimatedBoard>;
typedef Word = String;

class Game {
  Game({
    // Settings
    required this.gameSettings,
    required this.globalSettings,

    // State
    this.isRunning = false,
    this.isStarted = false,
    this.isFinished = false,
    this.animationInProgress = false,
    this.gettingSecretWord = false,

    // Base data
    required this.secretWord,
    required this.clue,

    // Game data
    required this.hiddenWord,
    required this.usedLetters,
    this.errorsCount = 0,
    this.isClueDisplayed = false,
  });

  // Settings
  final GameSettings gameSettings;
  final GlobalSettings globalSettings;

  // State
  bool isRunning;
  bool isStarted;
  bool isFinished;
  bool animationInProgress;
  bool gettingSecretWord;

  // Base data
  String secretWord;
  String clue;

  // Game data
  List<String> hiddenWord = [];
  List<String> usedLetters = [];
  int errorsCount;
  bool isClueDisplayed;

  factory Game.createNull() {
    return Game(
      // Settings
      gameSettings: GameSettings.createDefault(),
      globalSettings: GlobalSettings.createDefault(),
      // Base data
      secretWord: '',
      clue: '',
      // Game data
      hiddenWord: [],
      usedLetters: [],
    );
  }

  factory Game.createNew({
    required GameSettings? gameSettings,
    required GlobalSettings? globalSettings,
  }) {
    final GameSettings newGameSettings = gameSettings ?? GameSettings.createDefault();
    final GlobalSettings newGlobalSettings = globalSettings ?? GlobalSettings.createDefault();

    return Game(
      // Settings
      gameSettings: newGameSettings,
      globalSettings: newGlobalSettings,
      // State
      isRunning: true,
      gettingSecretWord: true,
      // Base data
      secretWord: '',
      clue: '',
      // Game data
      hiddenWord: [],
      usedLetters: [],
    );
  }

  String get hiddenWordAsString => hiddenWord.join();

  bool get canBeResumed => isStarted && !isFinished;
  bool get gameWon => isFinished && (hiddenWordAsString == secretWord);

  void dump() {
    printlog('');
    printlog('## Current game dump:');
    printlog('');
    printlog('$Game:');
    printlog('  Settings');
    gameSettings.dump();
    globalSettings.dump();
    printlog('  State');
    printlog('    isRunning: $isRunning');
    printlog('    isStarted: $isStarted');
    printlog('    isFinished: $isFinished');
    printlog('    animationInProgress: $animationInProgress');
    printlog('    gettingSecretWord: $gettingSecretWord');
    printlog('  Base data');
    printlog('    secretWord: $secretWord');
    printlog('    clue: $clue');
    printlog('  Game data');
    printlog('    hiddenWord: $hiddenWord');
    printlog('    usedLetters: $usedLetters');
    printlog('    errorsCount: $errorsCount');
    printlog('    isClueDisplayed: $isClueDisplayed');
    printlog('');
  }

  @override
  String toString() {
    return '$Game(${toJson()})';
  }

  Map<String, dynamic>? toJson() {
    return <String, dynamic>{
      // Settings
      'gameSettings': gameSettings.toJson(),
      'globalSettings': globalSettings.toJson(),
      // State
      'isRunning': isRunning,
      'isStarted': isStarted,
      'isFinished': isFinished,
      'animationInProgress': animationInProgress,
      'gettingSecretWord': gettingSecretWord,
      // Base data
      'secretWord': secretWord,
      'clue': clue,
      // Game data
      'hiddenWord': hiddenWord,
      'usedLetters': usedLetters,
      'errorsCount': errorsCount,
      'isClueDisplayed': isClueDisplayed,
    };
  }
}
